import torch
import torch.nn as nn
import time, copy, utils
from time import process_time, localtime, strftime
import sys, os, signal
import matplotlib.pyplot as plt
from torchsummary import summary
from utils import plot_confusion_matrix
import vars
import torchvision
from sklearn.metrics import precision_recall_fscore_support as score
from sklearn.metrics import confusion_matrix, cohen_kappa_score, precision_score, recall_score
import numpy as np
######################################################################
# Training the model
def train_model(model, criterion, optimizer, scheduler, num_epochs):
    since = time.time()
    train_loss = []
    val_loss = []
    train_acc = []
    val_acc = []
    best_model_wts = copy.deepcopy(model.state_dict())
    best_acc = 0.0

    #log_dir = '{}-{}-{}-batch-{}'.format(type(model).__name__, type(optimizer).__name__, vars.device, vars.batch_size)
    log_dir = '{}-{}-{}-batch-{}'.format(vars.model_name, type(optimizer).__name__, vars.device, vars.batch_size)
    if(vars.pre_trained):
        log_dir = log_dir + '-pretrained'
    if not os.path.exists(log_dir):
        os.mkdir(log_dir)

    log_file = open(log_dir+"\\train-{}.log".format(strftime("%Y-%m-%d %H-%M", localtime())),"w")
    log_file.write(str(model) + "\n" + '-' * 80)
    #log_file.write(str(summary(model, input_size=(3, vars.input_size, vars.input_size), batch_size=-1, device=vars.device.type)))
    log_file.write("\nOptimizer: \n" + str(optimizer) + "\n" + '-' * 80)
    log_file.write("\nscheduler: \n" + str(scheduler) + "\n" + '-' * 80)
    log_file.flush()
    for epoch in range(1, vars.num_epochs+1):
        if(vars.stop_training):
            break
        ########################################################################################
        #  # در هرلایه، کرنلها با هم مقایسه می شوند. درصورت صفر بودن فاصله ی آنها، در آموزش بعدی، تعداد کمتری کرنل درز شبکه استفاده میکنیم
        # for it1 in range(model.conv1.weight.size()[0]):
        #     a=model.conv1.weight[it1,:,:,:]                 
        #     for it2 in range(model.conv1.weight.size()[0]):  
        #         if it1!=it2:              
        #             b=model.conv1.weight[it2,:,:,:]
        #             c=torch.dist(a, b)
        #             print(c.cpu().detach().numpy())
                    
                    
        ######################################################################################           
        log_file.write('\n' + '-' * 80)
        log_str = '\nEpoch {}/{} --'.format(epoch, vars.num_epochs) + strftime("%Y-%m-%d %H:%M:%S", localtime())
        log_file.write(log_str) 
        print(log_str)
        print('-' * 10)


        for phase in ['train', 'val']:

         # Each epoch has a training and validation phase    
            if phase == 'train':
                model.train()  # Set model to training mode
            else:
                model.eval()   # Set model to evaluate mode
                
            running_loss = 0.0
            running_corrects = 0
            num_batchs = vars.dataset_sizes[phase] // vars.batch_size
            # Iterate over data.
            cur_batch = 0
            for inputs, labels in vars.dataloaders[phase]:
                inputs = inputs.to(vars.device)
                labels = labels.to(vars.device)

                cur_batch += 1                

                # zero the parameter gradients
                optimizer.zero_grad()

                # forward
                # track history if only in train
                
                with torch.set_grad_enabled(phase == 'train'):
                    if(vars.model_name == 'inception' and phase == 'train'):
                        outputs, aux_outputs = model(inputs)
                        loss1 = criterion(outputs, labels)
                        loss2 = criterion(aux_outputs, labels)
                        loss = loss1 + 0.4*loss2
                    else:
                        outputs = model(inputs)
                        loss = criterion(outputs, labels)
                        
                    _, preds = torch.max(outputs, 1)

                    # backward + optimize only if in training phase
                    if phase == 'train':
                        loss.backward()
                        optimizer.step()

                # statistics
                batch_correct = torch.sum(preds == labels.data)
                running_loss += loss.item() * inputs.size(0)
                running_corrects += batch_correct

                if(cur_batch % 10 == 0):
                    if(phase == 'train'):
                        log_str = '\r|{:2d}/{:2d}| {:4d}/{:4d}| - {:5s} loss {:6.4f} - acc {:6.2f}'.format(epoch, vars.num_epochs, cur_batch, num_batchs, phase, loss, batch_correct.item()*100 / inputs.size(0))
                        log_file.write(log_str) 
                        log_file.flush()
                        print(log_str, end="")
                    else:
                        print('\rval batch {}/{}\r'.format(cur_batch, num_batchs), end="")

                # if phase == 'train':
                #     train_loss.append(loss.item())
                #     train_acc.append(batch_correct.item()*100 / inputs.size(0))
                # else:
                #     val_loss.append(loss)
                #     val_acc.append(batch_correct.item()*100 / inputs.size(0))
                # utils.plot_graphs(train_loss, val_loss, train_acc, val_acc)
            epoch_loss = running_loss / vars.dataset_sizes[phase]
            epoch_acc = running_corrects.double() / vars.dataset_sizes[phase]
            
            if phase == 'train':
                train_loss.append(epoch_loss)
                train_acc.append(epoch_acc)
                if(scheduler._last_lr[0] > vars.min_learning_rate):
                    scheduler.step()
                print("\nnew lr: {}".format(scheduler._last_lr), end="")
                log_file.write("\nnew lr: {}".format(scheduler._last_lr)) 
            else:
                val_loss.append(epoch_loss)
                val_acc.append(epoch_acc)
                utils.plot_graphs(train_loss, val_loss, train_acc, val_acc)


            log_str = '\n{:5s} Loss: {:.4f} Acc: {:.2f}\n'.format(phase, epoch_loss, 100*epoch_acc)
            log_file.write(log_str) 
            print(log_str, end="")
            log_file.flush()

            # deep copy the model
            #if phase == 'val' and epoch_acc > best_acc:
            if (epoch_acc > best_acc) or (epoch % 5 == 0):
                best_acc = epoch_acc
                best_model_wts = copy.deepcopy(model.state_dict()) # model.state_dict().copy()                
                torch.save(best_model_wts, '{}\\ep{}-acc{:.2f}-loss{:.4f}.pth'.format(log_dir, epoch, best_acc*100, epoch_loss))

        print()

    time_elapsed = time.time() - since

    log_str = '\nTraining complete in {:.0f}m {:.0f}s\n'.format(time_elapsed // 60, time_elapsed % 60)
    log_file.write(log_str) 
    print(log_str, end="")
            
    
    log_str = '\nBest val Acc: {:4f}'.format(best_acc)
    log_file.write(log_str) 
    print(log_str, end="")
    log_file.close() 
    utils.plot_graphs(train_loss, val_loss, train_acc, val_acc, False, log_dir)
    
    # load best model weights
    model.load_state_dict(best_model_wts)
    return model



######################################################################
def test_model(model, criterion, dataset='test', n_batch = 10000):
    
    # for it1 in range(model.conv5.weight.size()[0]):
    #     a=model.conv5.weight[it1,:,:,:]                 
    #     for it2 in range(model.conv5.weight.size()[0]):  
    #         if it1!=it2:              
    #             b=model.conv5.weight[it2,:,:,:]
    #             c=torch.dist(a, b)
    #             print(c.cpu().detach().numpy())
    
    
    
    
    log_str = '\n' + '-'*80 + '\nTest on ' + vars.device.type
    since = time.time()
     
    model.eval()   # Set model to evaluate mode
    running_loss = 0.0
    running_corrects = 0

    total_test_time = 0
    # Iterate over data.
    cur_batch = 0
    total_samples = 0
    p_reds = []
    p_labs = []
    for inputs, labels in vars.dataloaders[dataset]:
        # Make a grid from batch
        #out = torchvision.utils.make_grid(inputs)

        #utils.imshow(out, title=[vars.class_names[x] for x in labels])
        inputs = inputs.to(vars.device)
        labels = labels.to(vars.device)

        cur_batch += 1                
        if(cur_batch > n_batch):
            break

        total_samples += inputs.size(0)   
        # forward
        with torch.set_grad_enabled(False):
            t1 = time.time()
            outputs = model(inputs)
            _, preds = torch.max(outputs, 1)
            
            total_test_time += time.time() - t1
            loss = criterion(outputs, labels)

        # statistics
        batch_correct = torch.sum(preds == labels.data)
        running_loss += loss.item() * inputs.size(0)
        running_corrects += batch_correct   
        p_reds[(cur_batch-1)*vars.batch_size:(cur_batch)*vars.batch_size-1] = preds.cpu().detach().numpy()
        p_labs[(cur_batch-1)*vars.batch_size:(cur_batch)*vars.batch_size-1] = labels.data .cpu().detach().numpy()
        
        #plot_confusion_matrix(labels.data, preds, normalize=False)

    loss = running_loss / total_samples
    acc = running_corrects.double() / total_samples
    
    
    time_elapsed = time.time() - since
    
    #confusion matrix
    cm = confusion_matrix(np.asarray(p_labs), np.asarray(p_reds))
    #cohen kappa score
    cks = cohen_kappa_score(np.asarray(p_labs), np.asarray(p_reds))
    #precision score
    ps = precision_score(np.asarray(p_labs), np.asarray(p_reds), average = None)
    #recal score
    rs = recall_score(np.asarray(p_labs), np.asarray(p_reds), average = None)
    
    metrics = {}
    metrics['confusion matrix'] = cm
    metrics['cohen kappa score'] = cks
    metrics['precision score'] = ps
    metrics['recall score'] = rs
    metrics['mean recall'] = np.mean(rs)
    metrics['mean precision'] = np.mean(ps)
    metrics['accuracy'] = np.diagonal(cm).sum() / float(len(np.asarray(p_labs)))
    print(metrics)
    #precision, recall, fscore, support = score(np.asarray(p_reds), np.asarray(p_labs))
    #accuracy, precision, recall, F1_score = utils.F_score(p_reds, p_labs)

    log_str += '\nLoss: {:.4f} Acc: {:.2f}'.format(loss, 100*acc)

    log_str += '\nTest completed over {} samples in {:.2f}s'.format(total_samples, time_elapsed)
    log_str += '\nPrediction time per sample {:.3f}ms'.format(time_elapsed * 1000 / total_samples)

    log_str += '\n(Excluding Data Load) Test completed over {} samples in {:.2f}s'.format(total_samples, total_test_time)
    log_str += '\n(Excluding Data Load) Prediction time per sample {:.3f}ms'.format(total_test_time * 1000 / total_samples)
    
    print(log_str, end="")
    return log_str

######################################################################
def test_model_metric(model, criterion, dataset='test', n_batch = 10000):
    
    log_str = '\n' + '-'*80 + '\nTest on ' + vars.device.type
    since = time.time()
     
    model.eval()   # Set model to evaluate mode

    total_test_time = 0
    # Iterate over data.
    cur_batch = 0
    total_samples = 0

    csv_file = open("mio-tcd-{}.log".format(strftime("%Y-%m-%d %H-%M", localtime())),"w")

    for inputs, labels in vars.dataloaders[dataset]:
        inputs = inputs.to(vars.device)

        cur_batch += 1                
        if(cur_batch > n_batch):
            break

        total_samples += inputs.size(0)   
        # forward
        with torch.set_grad_enabled(False):
            t1 = time.time()
            outputs = model(inputs)
            _, preds = torch.max(outputs, 1)
            for s, p in zip(labels, preds):
                csv_file.write(s)
                csv_file.write(',')
                csv_file.write(vars.class_names[p])
                csv_file.write('\n')
            csv_file.flush()

            total_test_time += time.time() - t1

    
    time_elapsed = time.time() - since

    log_str += '\nTest completed over {} samples in {:.2f}s'.format(total_samples, time_elapsed)
    log_str += '\nPrediction time per sample {:.3f}ms'.format(time_elapsed * 1000 / total_samples)

    log_str += '\n(Excluding Data Load) Test completed over {} samples in {:.2f}s'.format(total_samples, total_test_time)
    log_str += '\n(Excluding Data Load) Prediction time per sample {:.3f}ms'.format(total_test_time * 1000 / total_samples)
    
    print(log_str, end="")
    return log_str

######################################################################
def test_model_mio_tcd(model, criterion, dataset='test', n_batch = 10000):
    
    log_str = '\n' + '-'*80 + '\nTest on ' + vars.device.type
    since = time.time()
     
    model.eval()   # Set model to evaluate mode

    total_test_time = 0
    # Iterate over data.
    cur_batch = 0
    total_samples = 0

    csv_file = open("mio-tcd-{}.log".format(strftime("%Y-%m-%d %H-%M", localtime())),"w")

    for inputs, labels in vars.dataloaders[dataset]:
        inputs = inputs.to(vars.device)

        cur_batch += 1                
        if(cur_batch > n_batch):
            break

        total_samples += inputs.size(0)   
        # forward
        with torch.set_grad_enabled(False):
            t1 = time.time()
            outputs = model(inputs)
            _, preds = torch.max(outputs, 1)
            for s, p in zip(labels, preds):
                csv_file.write(s)
                csv_file.write(',')
                csv_file.write(vars.class_names[p])
                csv_file.write('\n')
            csv_file.flush()

            total_test_time += time.time() - t1

    
    time_elapsed = time.time() - since

    log_str += '\nTest completed over {} samples in {:.2f}s'.format(total_samples, time_elapsed)
    log_str += '\nPrediction time per sample {:.3f}ms'.format(time_elapsed * 1000 / total_samples)

    log_str += '\n(Excluding Data Load) Test completed over {} samples in {:.2f}s'.format(total_samples, total_test_time)
    log_str += '\n(Excluding Data Load) Prediction time per sample {:.3f}ms'.format(total_test_time * 1000 / total_samples)
    
    print(log_str, end="")
    return log_str
######################################################################
# Generic function to display predictions for a few images
def visualize_model(model, num_images=6):
    was_training = model.training
    model.eval()
    images_so_far = 0
    fig = plt.figure()

    with torch.no_grad():
        for i, (inputs, labels) in enumerate(vars.dataloaders['val']):
            inputs = inputs.to(vars.device)
            labels = labels.to(vars.device)

            outputs = model(inputs)
            _, preds = torch.max(outputs, 1)

            for j in range(inputs.size()[0]):
                images_so_far += 1
                ax = plt.subplot(num_images//2, 2, images_so_far)
                ax.axis('off')
                ax.set_title('predicted: {}'.format(vars.class_names[preds[j]]))
                utils.imshow(inputs.cpu().data[j])

                if images_so_far == num_images:
                    model.train(mode=was_training)
                    return
        model.train(mode=was_training)




# def initialize_model(model_name, num_classes, feature_extract, use_pretrained=True):
#     # Initialize these variables which will be set in this if statement. Each of these
#     #   variables is model specific.
#     model_ft = None
#     input_size = 0

#     if model_name == "resnet":
#         """ Resnet18
#         """
#         model_ft = models.resnet18(pretrained=use_pretrained)
#         set_parameter_requires_grad(model_ft, feature_extract)
#         num_ftrs = model_ft.fc.in_features
#         model_ft.fc = nn.Linear(num_ftrs, num_classes)
#         input_size = 224

#     elif model_name == "alexnet":
#         """ Alexnet
#         """
#         model_ft = models.alexnet(pretrained=use_pretrained)
#         set_parameter_requires_grad(model_ft, feature_extract)
#         num_ftrs = model_ft.classifier[6].in_features
#         model_ft.classifier[6] = nn.Linear(num_ftrs,num_classes)
#         input_size = 224

#     elif model_name == "vgg":
#         """ VGG11_bn
#         """
#         model_ft = models.vgg11_bn(pretrained=use_pretrained)
#         set_parameter_requires_grad(model_ft, feature_extract)
#         num_ftrs = model_ft.classifier[6].in_features
#         model_ft.classifier[6] = nn.Linear(num_ftrs,num_classes)
#         input_size = 224

#     elif model_name == "squeezenet":
#         """ Squeezenet
#         """
#         model_ft = models.squeezenet1_0(pretrained=use_pretrained)
#         set_parameter_requires_grad(model_ft, feature_extract)
#         model_ft.classifier[1] = nn.Conv2d(512, num_classes, kernel_size=(1,1), stride=(1,1))
#         model_ft.num_classes = num_classes
#         input_size = 224

#     elif model_name == "densenet":
#         """ Densenet
#         """
#         model_ft = models.densenet121(pretrained=use_pretrained)
#         set_parameter_requires_grad(model_ft, feature_extract)
#         num_ftrs = model_ft.classifier.in_features
#         model_ft.classifier = nn.Linear(num_ftrs, num_classes)
#         input_size = 224

#     elif model_name == "inception":
#         """ Inception v3
#         Be careful, expects (299,299) sized images and has auxiliary output
#         """
#         model_ft = models.inception_v3(pretrained=use_pretrained)
#         set_parameter_requires_grad(model_ft, feature_extract)
#         # Handle the auxilary net
#         num_ftrs = model_ft.AuxLogits.fc.in_features
#         model_ft.AuxLogits.fc = nn.Linear(num_ftrs, num_classes)
#         # Handle the primary net
#         num_ftrs = model_ft.fc.in_features
#         model_ft.fc = nn.Linear(num_ftrs,num_classes)
#         input_size = 299

#     else:
#         print("Invalid model name, exiting...")
#         exit()

#     return model_ft, input_size

# # Initialize the model for this run
# model_ft, input_size = initialize_model(model_name, num_classes, feature_extract, use_pretrained=True)

# # Print the model we just instantiated
# print(model_ft)        